import os
import random
import sys

import anndata as ad
import numpy as np
import pandas as pd
import torch
import torch.nn.modules.loss
import torch_geometric.transforms as T
from scipy import sparse
from sklearn.preprocessing import normalize
from torch import Tensor, nn
from torch.nn.functional import binary_cross_entropy, kl_div, mse_loss
from torch.utils.data import DataLoader
from torch_geometric.data import Data
from torch_geometric.nn import VGAE, GCNConv, InnerProductDecoder
from tqdm import tqdm

from TRAPT.Tools import RPMatrix


def seed_torch(seed=2023):
    random.seed(seed)
    os.environ['PYTHONHASHSEED'] = str(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)


class InnerProductDecoderWeight(InnerProductDecoder):
    def __init__(self, A_e, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.A_e = A_e

    def forward(
        self, z: Tensor, edge_index: Tensor = None, sigmoid: bool = True
    ) -> Tensor:
        adj = torch.matmul(z, z.t())
        return torch.sigmoid(adj) if sigmoid else adj


class VariationalGCNEncoder(torch.nn.Module):
    def __init__(self, in_channels, h_dim, z_dim):
        super(VariationalGCNEncoder, self).__init__()
        self.conv_h = GCNConv(in_channels, h_dim, cached=True)
        self.conv_mu = GCNConv(h_dim, z_dim, cached=True)
        self.conv_logstd = GCNConv(h_dim, z_dim, cached=True)

    def predict_h(self, x, edge_index):
        h = self.conv_h(x, edge_index).relu()
        return h

    def forward(self, x, edge_index):
        h = self.conv_h(x, edge_index).relu()
        return self.conv_mu(h, edge_index), self.conv_logstd(h, edge_index)


class CVAE(nn.Module):
    def __init__(self, input_dim, condition_dim, h_dim, z_dim):
        super(CVAE, self).__init__()
        self.h_encoder = torch.nn.Linear(input_dim + condition_dim, h_dim)
        self.act = torch.nn.ReLU()
        self.dropout = torch.nn.Dropout()
        self.layer_mu = torch.nn.Linear(h_dim, z_dim)
        self.layer_logstd = torch.nn.Linear(h_dim, z_dim)
        self.decoder = torch.nn.Linear(z_dim + condition_dim, input_dim)

    def reparametrize(self, mu: Tensor, logstd: Tensor) -> Tensor:
        if self.training:
            return mu + torch.randn_like(logstd) * torch.exp(logstd)
        else:
            return mu

    def predict_h(self, x):
        h = self.act(self.h_encoder(x))
        return h

    def kl_div(self):
        return torch.mean(
            0.5
            * torch.sum(
                torch.exp(self.logstd) + self.mu**2 - 1.0 - self.logstd, dim=1
            )
        )

    def forward(self, x):
        condition = x[:, -2:]
        h = self.act(self.h_encoder(x))
        self.mu = self.layer_mu(h)
        self.logstd = self.layer_logstd(h)
        z = self.reparametrize(self.mu, self.logstd)
        decoded = self.decoder(torch.concat([z, condition], dim=1))
        return decoded


class CalcSTM:
    def __init__(self, RP_Matrix, type, checkpoint_path, device='cpu'):
        self.type = type
        assert type in ['H3K27ac', 'ATAC']
        self.checkpoint_path = checkpoint_path
        self.device = device
        print(f"Using {self.device}")
        self.RP_Matrix = RP_Matrix
        self.data = torch.concat(
            [
                torch.tensor(self.RP_Matrix.TR.to_df().values),
                torch.tensor(self.RP_Matrix.Sample.to_df().values),
            ],
            dim=0,
        ).to(self.device)
        self.is_val = 0

    @staticmethod
    def get_cos_similar_matrix(m1, m2):
        num = m1.dot(m2.T)
        denom = np.linalg.norm(m1, axis=1).reshape(-1, 1) * np.linalg.norm(m2, axis=1)
        res = num / denom
        res[np.isneginf(res)] = 0
        return res

    def get_edge_index(self, A, B, n=10):
        self.A_s = 0
        self.A_e = A.shape[0]
        self.B_s = self.A_e
        self.B_e = self.B_s + B.shape[0]
        adj = np.zeros((self.B_e, self.B_e))
        AB_cos_sim = self.get_cos_similar_matrix(A.to_df().values, B.to_df().values)
        AB = (AB_cos_sim >= np.sort(AB_cos_sim, axis=1)[:, -n].reshape(-1, 1)).astype(
            int
        )
        adj[self.A_s : self.A_e, self.B_s : self.B_e] = AB
        adj[self.B_s : self.B_e, self.A_s : self.A_e] = AB.T
        row, col, value = sparse.find(adj)
        return torch.tensor(np.array([row, col]), dtype=torch.long)

    @staticmethod
    def sparse_to_tensor(data, type='sparse'):
        if type == 'dense':
            return torch.tensor(data.to_df().values)
        else:
            row, col, value = sparse.find(data)
            indices = torch.tensor(np.array([row, col]), dtype=torch.long)
            values = torch.tensor(value, dtype=torch.float)
            size = torch.Size(data.shape)
            return torch.sparse_coo_tensor(indices, values, size)

    def test(self, data):
        self.model_vgae.eval()
        with torch.no_grad():
            from sklearn.metrics import average_precision_score, roc_auc_score

            z = self.model_vgae.encode(
                data.x.to(self.device), data.edge_index.to(self.device)
            )
            y_pred = self.model_vgae.decoder(z, sigmoid=True)
            y_pred = y_pred.detach().cpu().numpy().flatten()
            index = np.random.choice(np.arange(len(self.y)), len(self.y) // 1000)
            return roc_auc_score(self.y[index], y_pred[index]), average_precision_score(
                self.y, y_pred
            )

    def save_graph(self):
        print(self.test(self.data))
        x, edge_index = self.data.x.to(self.device), self.data.edge_index.to(
            self.device
        )
        z = self.model_vgae.encode(x, edge_index).cpu().detach().numpy()
        A_pred = z.dot(z.T)
        TR_info = pd.DataFrame(list(self.RP_Matrix.TR.obs_names))
        TR_info['type'] = 'TR'
        Sample_info = pd.DataFrame(list(self.RP_Matrix.Sample.obs_names))
        Sample_info['type'] = self.type
        info = pd.concat([TR_info, Sample_info]).astype(str)
        info.columns = ['index', 'type']
        info = info.set_index('index')
        data = ad.AnnData(A_pred, obs=info, var=info)
        data.write_h5ad(f'{self.checkpoint_path}/A_pred_{self.type}.h5ad')

    def recon_loss(self, z):
        y_pred = self.model_vgae.decoder(z, sigmoid=True)
        loss = (
            self.norm
            * binary_cross_entropy(y_pred.view(-1), self.y, weight=self.weight).mean()
        )
        return loss

    def init_vgae(self, h, use_dl):
        seed_torch()
        self.h = h
        self.edge_index = self.get_edge_index(self.RP_Matrix.TR, self.RP_Matrix.Sample)
        print("Edge sum:", self.edge_index.shape[1])
        self.data = Data(x=self.data, edge_index=self.edge_index)
        if self.is_val:
            self.transform = T.RandomLinkSplit(
                num_val=0.2,
                num_test=0,
                neg_sampling_ratio=0,
                add_negative_train_samples=False,
            )
            self.train_data, self.val_data, self.test_data = self.transform(self.data)
            x = self.train_data.x.to(self.device)
            edge_index = self.train_data.edge_index.to(self.device)
        else:
            x = self.data.x.to(self.device)
            edge_index = self.data.edge_index.to(self.device)
        self.num_features = self.data.num_features
        self.num_nodes = self.data.num_nodes
        values = torch.ones(self.edge_index.shape[1])
        size = torch.Size([self.num_nodes] * 2)
        y = torch.sparse_coo_tensor(self.edge_index, values, size)
        y = y.to_dense().view(-1)
        pos_weight = (self.num_nodes**2 - y.sum()) / y.sum()
        weight = torch.ones_like(y)
        weight[y.numpy().astype(bool)] = pos_weight
        self.weight = weight.to(self.device)
        self.norm = self.num_nodes**2 / ((self.num_nodes**2 - y.sum()) * 2)
        self.y = y.to(self.device)

        encoder = VariationalGCNEncoder(self.data.num_features, self.h_dim, self.z_dim)
        decoder = InnerProductDecoderWeight(self.A_e)
        self.model_vgae = VGAE(encoder, decoder).to(self.device)
        self.optimizer_vgae = torch.optim.Adam(self.model_vgae.parameters(), lr=0.01)
        if (
            os.path.exists(f'{self.checkpoint_path}/model_student_{self.type}.pt')
            and not self.is_val
        ):
            checkpoint = torch.load(
                f'{self.checkpoint_path}/model_student_{self.type}.pt',
                map_location=self.device,
            )
            self.model_vgae.load_state_dict(checkpoint['model_state_dict'])
            self.optimizer_vgae.load_state_dict(checkpoint['optimizer_state_dict'])
        else:
            print('# Training ...')
            for epoch in range(self.epochs_vgae):
                self.model_vgae.train()
                self.optimizer_vgae.zero_grad()
                h = self.model_vgae.encoder.predict_h(x, edge_index)
                z = self.model_vgae.encode(x, edge_index)
                if use_dl:
                    dl_loss = mse_loss(h, self.h)
                else:
                    dl_loss = 0
                re_loss = self.recon_loss(z)
                kl_loss = (1 / self.num_nodes) * self.model_vgae.kl_loss()
                loss = re_loss + kl_loss + dl_loss
                loss.backward()
                self.optimizer_vgae.step()
                if epoch % 1 == 0:
                    if self.is_val:
                        auc, ap = self.test(self.val_data)
                        print(f"batch:", auc, ap)
                    print(
                        f"epoch: {epoch}, loss: {loss}, dl_loss: {dl_loss}, re_loss: {re_loss}, kl_loss: {kl_loss}"
                    )
                    if not self.is_val:
                        checkpoint = {
                            'model_state_dict': self.model_vgae.state_dict(),
                            'optimizer_state_dict': self.optimizer_vgae.state_dict(),
                        }
                        torch.save(
                            checkpoint,
                            f'{self.checkpoint_path}/model_student_{self.type}.pt',
                        )
            if not self.is_val:
                torch.save(
                    checkpoint,
                    f'{self.checkpoint_path}/model_student_{self.type}.pt',
                )
        if self.is_val:
            auc, ap = self.test(self.val_data)
            print(f"batch:", auc, ap)
        else:
            self.save_graph()

    def init_cvae(self):
        seed_torch()
        condition = np.ones(self.data.shape[0])
        condition[: self.RP_Matrix.TR.shape[0]] = 0
        condition = np.eye(2)[condition.astype(int)]
        condition = normalize(condition, 'l1', axis=0)
        condition = torch.tensor(condition, dtype=torch.float32).to(self.device)
        dataset = torch.concat([self.data, condition], dim=1)
        batch_size = 32
        train_loader = DataLoader(
            dataset=dataset,
            batch_size=batch_size,
            shuffle=True,
        )
        self.model_cvae = CVAE(
            self.data.shape[1], condition.shape[1], self.h_dim, self.z_dim
        ).to(self.device)
        self.optimizer_cvae = torch.optim.Adam(self.model_cvae.parameters(), lr=0.01)
        loss_func = nn.MSELoss()
        if (
            os.path.exists(f'{self.checkpoint_path}/model_teacher_{self.type}.pt')
            and not self.is_val
        ):
            checkpoint = torch.load(
                f'{self.checkpoint_path}/model_teacher_{self.type}.pt',
                map_location=self.device,
            )
            self.model_cvae.load_state_dict(checkpoint['model_state_dict'])
            self.optimizer_cvae.load_state_dict(checkpoint['optimizer_state_dict'])
        else:
            print('# Training ...')
            for epoch in range(self.epochs_cvae):
                for batch in tqdm(train_loader):
                    self.model_cvae.train()
                    self.optimizer_cvae.zero_grad()
                    decoded = self.model_cvae(batch)
                    kl_loss = self.model_cvae.kl_div() / batch_size
                    re_loss = (
                        loss_func(decoded.view(-1), batch[:, :-2].reshape(-1))
                        / batch_size
                    )
                    loss = re_loss + kl_loss
                    loss.backward()
                    self.optimizer_cvae.step()
                print(
                    f'epoch: {epoch}, loss: {loss}, kl_loss: {kl_loss}, re_loss: {re_loss}'
                )
                if not self.is_val:
                    checkpoint = {
                        'model_state_dict': self.model_cvae.state_dict(),
                        'optimizer_state_dict': self.optimizer_cvae.state_dict(),
                    }
                    torch.save(
                        checkpoint,
                        f'{self.checkpoint_path}/model_teacher_{self.type}.pt',
                    )

        return self.model_cvae.predict_h(dataset).detach()

    def run(self,use_dl=True):
        self.epochs_cvae = 10
        self.epochs_vgae = 100
        # self.epochs_cvae = 1
        # self.epochs_vgae = 1
        self.h_dim = 48
        self.z_dim = 24
        # Teacher
        z = self.init_cvae()
        # Student
        self.init_vgae(z,use_dl)


if __name__ == '__main__':
    type = sys.argv[1]
    library = sys.argv[2]
    use_dl = True
    checkpoint_path = library
    if len(sys.argv) > 3:
        use_dl = False if sys.argv[3].lower() == "false" else True
    if len(sys.argv) > 4:
        checkpoint_path = sys.argv[4]

    assert type in ['H3K27ac', 'ATAC']

    class RP_Matrix:
        TR = RPMatrix(library, 'RP_Matrix_TR.h5ad').norm('l1').get_data()
        Sample = RPMatrix(library, f'RP_Matrix_{type}.h5ad').norm('l1').get_data()

    CalcSTM(RP_Matrix, type, checkpoint_path).run(use_dl)



import random

import numpy as np
import pandas as pd
import tensorflow as tf
from tensorflow.keras import backend as K
from keras import Input, Model
from keras.callbacks import EarlyStopping
from keras.layers import Dense, Dropout, Layer
from keras.models import Model
from keras.optimizers import Adam
from keras.regularizers import Regularizer

def seed_tensorflow(seed=2023):
    random.seed(seed)
    np.random.seed(seed)
    tf.random.set_seed(seed)


# def custom_sigmoid(x, t=1):
#     return 1 / (1 + K.exp(-x / t))

class CustomSigmoid(Layer):
    def __init__(self, *, activity_regularizer=None, trainable=True, dtype=None, autocast=True, name=None, t=1, **kwargs):
        super().__init__(activity_regularizer=activity_regularizer, trainable=trainable, dtype=dtype, autocast=autocast, name=name, **kwargs)
        self.t = t
    def call(self, x):
        return 1 / (1 + K.exp(-x / self.t))

def custom_sigmoid(x, t=1):
    return CustomSigmoid(t=t)(x)


class SparseGroupLasso(Regularizer):
    """
    https://keras-ssg-lasso.readthedocs.io/en/latest/
    argmin(w) (1/2n * ||y - Xw||^2 + λ1 * ||w||1 + λ2 * ∑g∈G √pg * ||w_g||2)
    """

    def __init__(self, l1=0.01, l2=0.01, groups=None):
        self.l1 = l1
        self.l2 = l2
        groups = np.array(groups).astype("int32")
        self.p_l = K.variable(np.sqrt(np.bincount(groups)).reshape((1, -1)))
        self.groups = K.variable(groups, "int32")

    def __call__(self, x):
        regularization = 0.0
        if self.l1:
            regularization += self.l1 * K.sum(K.abs(x))
        if self.l2:
            regularization += self.l2 * K.sum(
                K.dot(
                    self.p_l,
                    K.sqrt(
                        tf.math.unsorted_segment_sum(
                            K.square(x), self.groups, num_segments=self.p_l.shape[1]
                        )
                    ),
                )
            )
        return regularization

    def get_config(self):
        return {
            "name": self.__class__.__name__,
            "l1": self.l1,
            "l2": self.l2,
            "groups": self.groups,
        }


class FeatureSelection:
    def __init__(self, args, data_ad, type):
        self.args = args
        self.data_ad = data_ad
        self.type = type
        self.epochs = 16
        self.batch_size = 32
        self.learning_rate = 0.001
        self.group_size = 10
        self.top = 10
        self.shuffle = True
        self.is_val = False
        self.early_stopping = EarlyStopping(patience=4)

    def TSFS(self, X, T):
        seed_tensorflow()
        pos_weight = float(len(T) - T.sum()) / T.sum()
        sample_weight = np.ones(len(T))
        sample_weight[T.astype(bool)] = pos_weight
        if self.args.use_dl:
            input = Input(X.shape[1:])
            y = Dense(2, activation="relu")(input)
            feature_extraction = Model(input, y)
            t = feature_extraction(input)
            output = Dense(1)(t)
            output = custom_sigmoid(output, 1)
            teacher = Model(input, output)
            teacher.compile(
                optimizer=Adam(self.learning_rate), loss="mse", weighted_metrics=[]
            )
            if self.is_val:
                teacher.fit(
                    X,
                    T,
                    epochs=self.epochs,
                    batch_size=self.batch_size,
                    callbacks=[self.early_stopping],
                    validation_split=0.15,
                    sample_weight=sample_weight,
                    shuffle=self.shuffle,
                    verbose=1,
                )
            else:
                teacher.fit(
                    X,
                    T,
                    epochs=self.epochs,
                    batch_size=self.batch_size,
                    sample_weight=sample_weight,
                    shuffle=self.shuffle,
                    verbose=0,
                )
            Y = feature_extraction.predict(X)
            seed_tensorflow()
            input = Input((X.shape[-1],))
            output = Dense(
                Y.shape[-1] * 10,
                activation="relu",
                kernel_regularizer=SparseGroupLasso(groups=self.groups),
            )(input)
        else:
            Y = np.expand_dims(T,-1)
            input = Input((X.shape[-1],))
            output = Dense(
                Y.shape[-1] * 10 * 2,
                activation="relu",
                kernel_regularizer=SparseGroupLasso(groups=self.groups),
            )(input)
        o_1 = Dense(Y.shape[-1], activation="relu")(output)
        student = Model(input, o_1)
        student.compile(optimizer=Adam(self.learning_rate), loss="mse")
        student.fit(
            X,
            Y,
            epochs=self.epochs,
            batch_size=self.batch_size,
            sample_weight=sample_weight,
            shuffle=self.shuffle,
            verbose=0,
        )
        weights = student.layers[1].get_weights()[0]
        weights = np.nan_to_num(weights, nan=0)
        weights = np.sum(np.square(weights), 1)
        return np.argsort(-weights),weights

    def train(self, X, y):
        seed_tensorflow()
        pos_weight = float(len(y) - y.sum()) / y.sum()
        sample_weight = np.ones(len(y))
        sample_weight[y.astype(bool)] = pos_weight
        input = Input(X.shape[1:])
        input = Dropout(0.1)(input)
        output = Dense(2, activation="relu")(input)
        output = Dense(1)(output)
        output = custom_sigmoid(output, 1)
        model = Model(input, output)
        model.compile(
            optimizer=Adam(self.learning_rate), loss="mse", weighted_metrics=[]
        )
        if self.is_val:
            model.fit(
                X,
                y,
                epochs=self.epochs,
                callbacks=[self.early_stopping],
                validation_split=0.15,
                batch_size=self.batch_size,
                sample_weight=sample_weight,
                shuffle=self.shuffle,
                verbose=1,
            )
        else:
            model.fit(
                X,
                y,
                epochs=self.epochs,
                batch_size=self.batch_size,
                sample_weight=sample_weight,
                shuffle=self.shuffle,
                verbose=0,
            )
        return model

    def get_corr(self, v1: list, v2: list):
        return np.array([np.corrcoef(v1, v2[i])[0, 1] for i in range(len(v2))])

    def sort_by_group(self, vec):
        vec = self.get_corr(vec, self.data_ad.X)
        self.index = np.argsort((vec.min() - vec))
        self.data_ad = self.data_ad[self.index]
        self.groups = [i // self.group_size for i in range(len(self.index))]

    def run(self):
        geneset = pd.read_csv(self.args.input, header=None)[0]
        genes = self.data_ad.var_names.values
        y = np.in1d(genes, geneset).astype(int)
        self.sort_by_group(y)
        X = self.data_ad.X.T
        true_index = np.where(y.astype(bool))[0]
        np.random.seed(2023)
        false_index = np.random.choice(
            np.where(~y.astype(bool))[0], self.args.background_genes, replace=False
        )
        index = np.append(true_index, false_index)
        feature_indexs,weights = self.TSFS(X[index], y[index])
        feature_indexs = feature_indexs[: self.top]
        info_samples = pd.DataFrame(weights[feature_indexs], index=self.data_ad.obs_names.values[feature_indexs])
        features = X[:, feature_indexs]
        model = self.train(features[index], y[index])
        gene_vec = model.predict(features)
        gene_vec = pd.DataFrame(gene_vec, index=self.data_ad.var_names)
        return gene_vec,info_samples



from concurrent.futures import ThreadPoolExecutor

import numpy as np
import pandas as pd
from numba import jit
from tqdm import tqdm


class CalcTRAUC:
    def __init__(self, args, RP_Matrix_TR_Sample, w):
        self.args = args
        self.matrix = RP_Matrix_TR_Sample
        self.matrix.X *= w
        self.genes = self.matrix.var_names
        self.geneset = pd.read_csv(self.args.input, header=None)[0]

    @staticmethod
    @jit(nopython=True, nogil=True)
    def get_auc(params):
        i, j, labels, vec = params
        l_p = labels > 0.5
        p_c = np.sum(l_p)
        n_c = len(l_p) - p_c
        if p_c > 1 and n_c > 1:
            index = np.argsort(vec)
            l_p_rank = np.where(l_p[index])[0]
            rank_sum = np.sum(l_p_rank)
            auc = (rank_sum - p_c * (1 + p_c) / 2) / (p_c * n_c)
        else:
            auc = 0
        return i, j, auc

    def iter_params(self, gene_vec, trunk):
        start = trunk * self.args.trunk_size
        end = (trunk + 1) * self.args.trunk_size
        tr_rp = self.matrix[start:end].to_df().values
        for i in range(tr_rp.shape[0]):
            tr_vec = tr_rp[i]
            yield start + i, 0, gene_vec, tr_vec

    def run(self):
        gene_vec = np.in1d(self.genes, self.geneset)

        auc = np.zeros((self.matrix.shape[0], 1))
        with ThreadPoolExecutor(self.args.threads) as pool:
            trunk_count = int(self.matrix.shape[0] / self.args.trunk_size) + 1
            for trunk in tqdm(range(trunk_count)):
                data = self.iter_params(gene_vec, trunk)
                tasks = [pool.submit(self.get_auc, params) for params in data]
                for task in tasks:
                    i, j, score = task.result()
                    auc[i, j] = score

        auc = pd.DataFrame(auc, index=self.matrix.obs_names)
        return auc



import os
import numpy as np
import pandas as pd
import argparse
import re

# from TRAPT.Tools import RP_Matrix, Args, Type
# from TRAPT.CalcTRAUC import CalcTRAUC
# from TRAPT.DLFS import FeatureSelection

from Tools import RP_Matrix, Args, Type
from CalcTRAUC import CalcTRAUC
from DLFS import FeatureSelection

def runTRAPT(args: Args):
    rp_matrix = RP_Matrix(args.library)
    obs = rp_matrix.TR.obs

    if os.path.exists(f"{args.output}/H3K27ac_RP.csv"):
        H3K27ac_RP = pd.read_csv(f"{args.output}/H3K27ac_RP.csv", header=None)[0]
    else:
        FS_H3K27ac = FeatureSelection(args, rp_matrix.H3K27ac, Type.H3K27ac)
        H3K27ac_RP,H3K27ac_info_samples = FS_H3K27ac.run()
        H3K27ac_RP.to_csv(f"{args.output}/H3K27ac_RP.csv", index=False, header=False)
        H3K27ac_info_samples.to_csv(f"{args.output}/H3K27ac_info_samples.csv", header=False)

    if os.path.exists(f"{args.output}/ATAC_RP.csv"):
        ATAC_RP = pd.read_csv(f"{args.output}/ATAC_RP.csv", header=None)[0]
    else:
        FS_ATAC = FeatureSelection(args, rp_matrix.ATAC, Type.ATAC)
        ATAC_RP,ATAC_info_samples = FS_ATAC.run()
        ATAC_RP.to_csv(f"{args.output}/ATAC_RP.csv", index=False, header=False)
        ATAC_info_samples.to_csv(f"{args.output}/ATAC_info_samples.csv", header=False)

    if os.path.exists(f"{args.output}/RP_TR_H3K27ac_auc.csv"):
        RP_TR_H3K27ac_auc = pd.read_csv(
            f"{args.output}/RP_TR_H3K27ac_auc.csv", index_col=0, header=None
        )
    else:
        H3K27ac_RP = H3K27ac_RP.values.flatten()
        CTR_TR = CalcTRAUC(args, rp_matrix.TR_H3K27ac, H3K27ac_RP)
        RP_TR_H3K27ac_auc = CTR_TR.run()
        RP_TR_H3K27ac_auc.to_csv(f"{args.output}/RP_TR_H3K27ac_auc.csv", header=False)

    if os.path.exists(f"{args.output}/RP_TR_ATAC_auc.csv"):
        RP_TR_ATAC_auc = pd.read_csv(
            f"{args.output}/RP_TR_ATAC_auc.csv", index_col=0, header=None
        )
    else:
        ATAC_RP = ATAC_RP.values.flatten()
        CTR_TR = CalcTRAUC(args, rp_matrix.TR_ATAC, ATAC_RP)
        RP_TR_ATAC_auc = CTR_TR.run()
        RP_TR_ATAC_auc.to_csv(f"{args.output}/RP_TR_ATAC_auc.csv", header=False)

    data_auc = pd.concat([RP_TR_H3K27ac_auc, RP_TR_ATAC_auc], axis=1)
    data_auc /= np.linalg.norm(data_auc, axis=0, keepdims=True)
    TR_activity = pd.DataFrame(
        np.sum(data_auc.values, axis=1), index=data_auc.index, columns=[1]
    )
    TR_detail = pd.concat([TR_activity, data_auc], axis=1).reset_index()
    TR_detail.columns = ["TR", "TR activity", "RP_TR_H3K27ac_auc", "RP_TR_ATAC_auc"]
    if args.tr_type != "all":
        TR_detail = TR_detail[TR_detail.TR.str.contains({
            "tf":"Sample",
            "tcof":"TcoF",
            "cr":"CR"
        }[args.tr_type], flags=re.IGNORECASE)]
    if args.source != "all":
        TR_detail = TR_detail[TR_detail.Source.str.contains(args.source, flags=re.IGNORECASE)]
    obs.index.name = "TR"
    TR_detail = TR_detail.merge(obs.reset_index(), on="TR").sort_values(
        "TR activity", ascending=False
    )
    TR_detail.to_csv(os.path.join(args.output, "TR_detail.txt"), index=False, sep="\t")
    return TR_detail

def str2bool(v):
    if isinstance(v, bool):
       return v
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')

def main():
    description = "usage: %prog [options] -l [LIBRARY] -i [INPUT] -o [OUTPUT]"
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument("--library", "-l", type=str, default="library", help = "Enter the library path, default is './library'")
    parser.add_argument("--threads", "-p", type=int, default=16, help = "Number of threads to launch, default is 16")
    parser.add_argument("--trunk_size", "-s", type=int, default=32768, help = "Block size. If the memory is insufficient, set a smaller value. The default value is 32768")
    parser.add_argument("--input", "-i", type=str, default=None, help = "Enter a gene list")
    parser.add_argument("--output", "-o", type=str, default=None, help = "Enter an output folder")
    parser.add_argument("--background_genes", "-b", type=int, default=6000, help = "Background gene count")
    parser.add_argument("--use_dl", "-d", type=str2bool, default=True, help = "Using knowledge distillation")
    parser.add_argument("--tr_type", type=str, default="all", help = "all/tr/tcof/cr")
    parser.add_argument("--source", type=str, default="all", help = "all/cistrome/chip_altas/gtrd/remap/chip-atlas/remap/encode/geo")
    args = Args(**dict(parser.parse_args()._get_kwargs()))
    os.environ["CUDA_VISIBLE_DEVICES"] = ""
    os.environ["TF_ENABLE_ONEDNN_OPTS"] = "0"
    runTRAPT(args)

if __name__ == "__main__":
    main()



import os
import re
import sys

from glob import glob
from multiprocessing import Pool

import anndata
import numpy as np
import pandas as pd
from scipy import sparse
from tqdm import tqdm


class CalcSampleRPMatrix:
    def __init__(self, library='library', output='library', type='H3K27ac'):
        self.library = library
        self.output = output
        self.type = type
        self.hg38_refseq = os.path.join(library, 'hg38_refseq.ucsc')
        self.dhs_hg38_rose = os.path.join(library, 'dhs_hg38_rose.bed')
        self.DHS_TO_GENE = os.path.join(library, 'dhs_hg38_rose_DHS_TO_GENE.txt')

    def dhs2gene(self, sample):
        try:
            r = 100000
            d_ = 10000
            m = 0.01
            alpha = (r - d_) / np.log(2 / m - 1)
            sample_name = re.findall(r'dhs@(.*?).csv', sample)[0]
            vec = pd.read_csv(sample, sep='\t', header=None)[0].values
            rp_vec = []
            for gene in self.genes:
                index = self.dhs_gene_g.loc[gene, 'index']
                s = vec[index]
                d = self.dhs_gene_g.loc[gene, 'DISTANCE']
                w = np.ones(d.shape)
                w[d > d_] = 2 / (np.exp((d[d > d_] - d_) / alpha) + 1)
                w[d > r] = 0
                rp = np.mean(np.multiply(w, s))
                rp_vec.append(rp)
            rp_vec = np.log(np.array(rp_vec) + 1)
            rp_vec = sparse.csr_matrix(rp_vec)
            return sample_name, rp_vec
        except:
            print('Error %s !' % sample)
            return None

    def run(self):
        ucsc_gene = pd.read_csv(self.hg38_refseq, sep='\t')
        ucsc_gene = ucsc_gene[['name', 'name2']]
        ucsc_gene.columns = ['GENES', 'SYMBOLS']
        ucsc_gene = ucsc_gene.drop_duplicates()
        dhs_hg38 = pd.read_csv(self.dhs_hg38_rose, sep='\t', header=None)
        dhs_hg38 = dhs_hg38.reset_index()[['index', 0]]
        dhs_hg38.columns = ['index', 'DHS']
        dhs_gene = pd.read_csv(self.DHS_TO_GENE, sep='\t')
        dhs_gene = dhs_gene.iloc[:, [0, 4, 5]]
        dhs_gene.columns = ['DHS', 'GENES', 'DISTANCE']
        dhs_gene_merge = dhs_gene.merge(dhs_hg38, on='DHS')
        dhs_gene_merge = dhs_gene_merge.merge(ucsc_gene, on='GENES')
        dhs_gene_merge = dhs_gene_merge.drop_duplicates(['DHS', 'SYMBOLS'])
        self.dhs_gene_g = dhs_gene_merge.groupby('SYMBOLS').agg(
            {'DISTANCE': list, 'index': list}
        )
        self.dhs_gene_g['DISTANCE'] = self.dhs_gene_g['DISTANCE'].apply(
            lambda x: np.array(x)
        )
        self.genes = self.dhs_gene_g.index
        rp_matrix = []
        samples = sorted(
            glob(os.path.join(self.library, f'{self.type}_matrix', 'dhs@*_hg38.csv'))
        )
        args = list(map(lambda sample: (sample), samples))
        rp_matrix = []
        sample_list = []
        with Pool(16) as pool:
            for row in tqdm(pool.imap(self.dhs2gene, args), total=len(args)):
                if row:
                    sample_name, rp_vec = row
                    sample_list.append(sample_name)
                    rp_matrix.append(rp_vec)
        rp_matrix = sparse.vstack(rp_matrix, dtype='float32')
        rp_matrix_ad = anndata.AnnData(rp_matrix)
        rp_matrix_ad.var_names = self.genes
        rp_matrix_ad.obs_names = sample_list
        rp_matrix_ad.write_h5ad(
            os.path.join(self.output, f'RP_Matrix_{self.type}.h5ad')
        )


if __name__ == '__main__':
    type = sys.argv[1]
    library = sys.argv[2]
    assert type in ['H3K27ac', 'ATAC']
    CalcSampleRPMatrix(library=library, output=library, type=type).run()



import os
import re
import sys

from multiprocessing import Pool

import anndata as ad
import h5py
import numpy as np
import pandas as pd
from scipy import sparse
from tqdm import tqdm


class CalcTRRPMatrix():
    def __init__(self,library='library',output='library'):
        self.library = library
        self.output = output
        self.hg38_refseq = os.path.join(library,'hg38_refseq.ucsc')
        self.dhs_hg38_rose = os.path.join(library,'dhs_hg38_rose.bed')
        self.DHS_TO_GENE = os.path.join(library,'dhs_hg38_rose_DHS_TO_GENE.txt')
    
    def dhs2gene(self,sample):
        id = re.findall('dhs@(.*?).csv',sample)[0]
        try:
            r = 100000
            d_ = 10000
            e = self.TR_info.loc[id,'Distal Intergenic Percentage']
            if e > 0:
                m = e
            else:
                m = 0.01
            alpha = (r-d_)/np.log(2/m-0.99)
            sample_name = re.findall(r'dhs@(.*?).csv',sample)[0]
            vec = pd.read_csv(sample,sep='\t',header=None)[0].values
            rp_vec = []
            for gene in self.genes:
                index = self.dhs_gene_g.loc[gene,'index']
                s = vec[index]
                d = self.dhs_gene_g.loc[gene,'DISTANCE']
                w = np.ones(d.shape)
                w[d > d_] = 2/(np.exp((d[d > d_]-d_)/alpha)+1)
                w[d > r] = 0
                rp = np.mean(np.multiply(w,s))
                rp_vec.append(rp)
            rp_vec = sparse.csr_matrix(rp_vec)
            return sample_name,rp_vec
        except:
            print('Error %s !' % sample)
            return None

    def run(self):
        ucsc_gene = pd.read_csv(self.hg38_refseq,sep='\t')
        ucsc_gene = ucsc_gene[['name','name2']]
        ucsc_gene.columns = ['GENES','SYMBOLS']
        ucsc_gene = ucsc_gene.drop_duplicates()
        dhs_hg38 = pd.read_csv(self.dhs_hg38_rose,sep='\t',header=None)
        dhs_hg38 = dhs_hg38.reset_index()[['index',0]]
        dhs_hg38.columns = ['index','DHS']
        dhs_gene = pd.read_csv(self.DHS_TO_GENE,sep='\t')
        dhs_gene = dhs_gene.iloc[:,[0,4,5]]
        dhs_gene.columns = ['DHS','GENES','DISTANCE']
        dhs_gene_merge = dhs_gene.merge(dhs_hg38,on='DHS')
        dhs_gene_merge = dhs_gene_merge.merge(ucsc_gene,on='GENES')
        dhs_gene_merge = dhs_gene_merge.drop_duplicates(['DHS','SYMBOLS'])
        self.dhs_gene_g = dhs_gene_merge.groupby('SYMBOLS').agg({
            'DISTANCE':list,
            'index': list
        })
        self.dhs_gene_g['DISTANCE'] = self.dhs_gene_g['DISTANCE'].apply(lambda x:np.array(x))
        self.genes = self.dhs_gene_g.index
        self.TR_info = pd.read_csv(os.path.join(self.library,'TRs_info.txt'),sep='\t',index_col=0)
        # New data
        # from glob import glob
        # samples = golb(os.path.join(self.library,'TR_matrix','dhs@*.csv'))
        tr_dhs_ad = h5py.File(os.path.join(self.library,'TR_DHS.h5ad'))
        samples = np.array(tr_dhs_ad['obs']['tr'],dtype=str)
        samples = list(map(lambda x:os.path.join(self.library,'TR_matrix',f'dhs@{x}.csv'),samples))

        args = list(map(lambda sample:(sample),samples))
        rp_matrix = []
        sample_list = []
        with Pool(16) as pool:
            for row in tqdm(pool.imap(self.dhs2gene, args),total=len(args)):
                if row:
                    sample_name,rp_vec = row
                    sample_list.append(sample_name)
                    rp_matrix.append(rp_vec)

        rp_matrix = sparse.vstack(rp_matrix,dtype='float32')
        rp_matrix_ad = ad.AnnData(rp_matrix)
        rp_matrix_ad.var_names = self.genes
        rp_matrix_ad.obs_names = sample_list

        obs = rp_matrix_ad.obs
        obs.index.name = 'tr'
        obs = obs.join(self.TR_info,how='left')
        obs['index'] = range(len(obs))

        rp_matrix_ad.obs = obs
        rp_matrix_ad.write_h5ad(os.path.join(self.output,f'RP_Matrix_TR.h5ad'))

if __name__ == '__main__':
    library = sys.argv[1]
    CalcTRRPMatrix(library=library,output=library).run()



import os

import anndata as ad
import numpy as np
from sklearn.preprocessing import MinMaxScaler, StandardScaler

class Args:
    def __init__(self, input, output, library="library", 
                 threads=16, trunk_size=2048 * 16, background_genes=6000, 
                 use_dl=True, tr_type="all", source="all") -> None:
        self.input = input
        self.output = output
        self.library = library
        self.threads = threads
        self.trunk_size = trunk_size
        self.background_genes = background_genes
        self.use_dl = use_dl
        self.tr_type = tr_type
        self.source = source
        if not os.path.exists(output):
            os.mkdir(output)

class Type:
    H3K27ac = 'H3K27ac'
    ATAC = 'ATAC'


class RPMatrix:
    def __init__(self, library, name, to_array=True):
        self.data = ad.read_h5ad(os.path.join(library, name))
        if to_array:
            self.data.X = self.data.to_df().values

    def norm(self, type="l2", axis=1):
        assert type in ["l1", "l2"]
        self.data.X = self.data.X - self.data.X.min()
        if type == "l1":
            self.data.X /= np.sum(self.data.X, axis=axis, keepdims=True).clip(min=1e-17)
        if type == "l2":
            self.data.X /= np.linalg.norm(self.data.X, axis=axis, keepdims=True).clip(
                min=1e-17
            )
        return self

    def standard_scale(self, axis=1):
        ss = StandardScaler()
        if axis == 0:
            self.data.X = ss.fit_transform(self.data.X)
        if axis == 1:
            self.data.X = ss.fit_transform(self.data.X.T).T
        return self

    def binarization(self):
        self.data.X = (self.data.X > 1e-17).astype(np.float32)
        return self

    def minmax_scale(self, axis=1):
        ss = MinMaxScaler()
        if axis == 0:
            self.data.X = ss.fit_transform(self.data.X)
        if axis == 1:
            self.data.X = ss.fit_transform(self.data.X.T).T
        return self

    def add(self, data):
        self.data.X += data.X
        return self

    def get_data(self):
        return self.data


class RP_Matrix:
    def __init__(self, library) -> None:
        self.TR = RPMatrix(library, 'RP_Matrix_TR.h5ad').norm().get_data()
        self.TR_H3K27ac = (
            RPMatrix(library, 'RP_Matrix_TR_H3K27ac.h5ad')
            .norm()
            .add(self.TR)
            .get_data()
        )
        self.TR_ATAC = (
            RPMatrix(library, 'RP_Matrix_TR_ATAC.h5ad').norm().add(self.TR).get_data()
        )
        self.H3K27ac = (
            RPMatrix(library, 'RP_Matrix_H3K27ac.h5ad').standard_scale().get_data()
        )
        self.ATAC = RPMatrix(library, 'RP_Matrix_ATAC.h5ad').standard_scale().get_data()




import os
import sys

import anndata as ad
import numpy as np


class CalcTRSampleRPMatrix:
    def __init__(self, library='library', output='library', type='H3K27ac'):
        self.data_ad_sample = ad.read_h5ad(
            os.path.join(library, f'RP_Matrix_{type}.h5ad')
        )
        self.data_ad_TR = ad.read_h5ad(os.path.join(library, 'RP_Matrix_TR.h5ad'))
        self.A = ad.read_h5ad(os.path.join(library, f'A_pred_{type}.h5ad'))
        self.A.X[np.diag_indices_from(self.A.X)] = 0
        i = self.data_ad_TR.shape[0]
        self.A = self.A[:i, i:]
        self.library = library
        self.output = output
        self.type = type

    def run(self):
        self.A = self.A.to_df().values
        self.A[self.A < 0.0] = 0
        self.X = self.data_ad_sample.to_df().values
        self.A = self.A / (np.sum(self.A, axis=1, keepdims=True) + 1e-17)
        self.X = self.X / (np.linalg.norm(self.X, axis=1, keepdims=True) + 1e-17)
        matrix = self.A.dot(self.X)
        data_ad = ad.AnnData(
            matrix, obs=self.data_ad_TR.obs, var=self.data_ad_TR.var, dtype='float32'
        )
        data_ad.write_h5ad(os.path.join(self.library, f'RP_Matrix_TR_{self.type}.h5ad'))


if __name__ == '__main__':
    type = sys.argv[1]
    library = sys.argv[2]
    assert type in ['H3K27ac', 'ATAC']
    print(f'\033[1;31m # # # CalcTRSampleRPMatrix # # #\033[0m')
    CalcTRSampleRPMatrix(library=library, output=library, type=type).run()
